/*
 * File         : CyclicRangeRep.C
 * Author       : Bryan Carpenter
 * Revision     : $Revision: 1.13 $
 * Updated      : $Date: 1999/11/05 22:29:38 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

// Operations on cyclic distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRangeRep': cyclic-distributed template range.          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* CyclicRangeRep :: shell() const {
  return new CollapsedRangeRep(numLayers) ;
}

void CyclicRangeRep :: location(Location* loc, const int glb) const {

  loc->sub = glb / numProcs ;
  loc->crd = glb % numProcs ;
  loc->dim = dimension ;
}

void CyclicRangeRep :: block(Block* blk, const int ker_glb) const {

  // `ker_glb' is a global subscript in the kernel range.

  blk->sub_stp = 1 ;
  blk->glb_stp = numProcs ;

  blk->sub_bas = 0 ;
  blk->glb_bas = ker_glb ;

  blk->count  = (extent - ker_glb + numProcs - 1) / numProcs ;
}

void CyclicRangeRep :: block2(Block* blk, const int ker_glb,
                              const int count, const int base) const {

  // `ker_glb' is a global subscript in the kernel range.

  // (Stripped down version of `block3'---can probably be simplified.)

  blk->sub_stp = 1 ;
  blk->glb_stp = numProcs ;

  const int disp = ker_glb - base ;
  int k_lo, k_hi ;

  const int a = disp ;
  const int b = numProcs ;

  if(a >= 0)
    k_lo = a - b * (a / b) ;
  else
    k_lo = a + b * ((-a + b - 1) / b) ;

  if(count > a)
    k_hi = a + b * ((count - 1 - a) / b) ;
  else
    k_hi = a - b * ((-count + a + b) / b) ;

  blk->sub_bas = (base + k_lo) / numProcs ;
  blk->count   = max((k_hi - k_lo) / numProcs + 1, 0) ;

  blk->glb_bas = k_lo ;
}

void CyclicRangeRep :: block3(Block* blk, const int ker_glb,
                              const int count, const int base,
                              const int stride) const {

  // `ker_glb' is a global subscript in the kernel range.

  // Compute `f', the GCD of `numProcs' and `stride', and `f_s',
  // a solution of
  //
  //   f_s * s = f  mod  p
  // 
  // (s = stride, p = numProcs)

  // (See for example "High Performance Compilers for Parallel Computing",
  // Michael Wolfe, section 4.2)

  int g, g_s ;

  int f   = numProcs ;
  int f_s = 0 ;
  if(stride > 0) {
    g   = stride ;
    g_s = 1 ;
  }
  else {
    g   = -stride ;
    g_s = -1 ;
  }
  while(g) {
    const int r = f / g ;

    const int temp1 = g ;
    g    = f - r * g ;
    f    = temp1 ;
  
    const int temp2 = g_s ;
    g_s  = f_s - r * g_s ;
    f_s  = temp2 ;
  }

  int k_step = numProcs / f ;

  // Define `k_lo' and `k_hi' for current value of `ker_glb'

  // Compute the smallest and largest solutions of
  //
  //     base + k * s = crd    mod  p
  //
  // in the range 0, ..., count - 1, viz
  //
  //     k_lo = a + b * ceiling(- a / b)
  //     k_hi = a + b * floor((count - 1 - a) / b)
  //
  // where a = c / f * f_s, b = p / f, c = crd - base.
  //
  // Solutions exist iff `f' divides `c'.

  blk->sub_stp = stride / f ;
  blk->glb_stp = k_step ;

  const int disp = ker_glb - base ;
  int k_lo, k_hi ;

  const int a = (disp / f) * f_s ;
  const int b = k_step ;

  if(a >= 0)
    k_lo = a - b * (a / b) ;
  else
    k_lo = a + b * ((-a + b - 1) / b) ;

  if(count > a)
    k_hi = a + b * ((count - 1 - a) / b) ;
  else
    k_hi = a - b * ((-count + a + b) / b) ;

  blk->sub_bas = (base + stride * k_lo) / numProcs ;
  blk->count   = max((k_hi - k_lo) / k_step + 1, 0) ;

  blk->glb_bas = k_lo ;
}

void CyclicRangeRep :: crds2(int* ker_count, int* ker_bas, int* ker_str,
                             const int count, const int base) const {
  *ker_count = kernel->size() ;
  *ker_bas   = 0 ;
  *ker_str   = 1 ;
}

void CyclicRangeRep :: crds3(int* ker_count, int* ker_bas, int* ker_str,
                             const int count, const int base,
                             const int stride) const {

  // Compute `f', the GCD of `numProcs' and `stride'.

  int f = numProcs ;
  int g = stride > 0 ? stride : - stride ;

  while(g) {
    const int temp = g ;
    g = f % g ;
    f = temp ;
  }

  *ker_count = kernel->size() / f ;
  *ker_bas   = base % f ;
  *ker_str   = f ;
}

int CyclicRangeRep :: volume() const {return numLayers ; }

int CyclicRangeRep :: glb(const int sub, const int ker_glb) const {
  return ker_glb + numProcs * sub ;
}

int CyclicRangeRep :: isAligned(RangeRep* x) const {
  if(x == (RangeRep*) this) return 1 ;
  if(x->format() == DIST_CYCLIC && x->dim() == dimension) {
    CyclicRangeRep* t = (CyclicRangeRep*) x->tem() ;
    return t->numLayers == numLayers && t->extent == extent &&
           x->size() == extent && x->str() == 1 ;
  }
  else
    return 0 ;
}

void CyclicRangeRep :: show() {
  AdlibPrintf("CyclicRangeRep :: show: size = %d ker = %p lev = %d\n",
              extent, kernel, level) ; 
}

MEMORY_POOL_DEFN(CyclicRangeRep)


// ------------------------------------------------------------------
// Things to do
//
//   In `block3', maybe can solve directly for `a' rather than
//   constructing `f_s'?
//
//   `block2' can probably be simplified.

